import os, re
import pyperclip
import pprint
import argparse
import json
from r3dir.encoder import encode, decode

hackvertor_tags_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'hackvertor')

hackvertor_encoder = os.path.join(hackvertor_tags_folder, 'encoder.js')

tags = [
    {
        "argument1Default": "302",
        "argument2Default": "",
        "code": 'const mainDomain = "[[MAIN_DOMAIN]]";\nconst httpsEnforced = false;\noutput = encode(input, statusCode, mainDomain, ignorePart, httpsEnforced, true);',
        "argument1Type": "Number",
        "numberOfArgs": 2,
        "argument1": "statusCode",
        "argument2Type": "String",
        "argument2": "ignorePart",
        "language": "JavaScript",
        "tagName": "_r3dir_encode"
    },
    {
        "code": 'const mainDomain = "[[MAIN_DOMAIN]]";\noutput = decode(input, mainDomain);',
        "numberOfArgs": 0,
        "language": "JavaScript",
        "tagName": "_r3dir_decode"
    },
    {
        "argument1Default": "302",
        "argument2Default": "",
        "code": 'const mainDomain = "[[MAIN_DOMAIN]]";\nconst httpsEnforced = true;\noutput = encode(input, statusCode, mainDomain, ignorePart, httpsEnforced, true);',
        "argument1Type": "Number",
        "numberOfArgs": 2,
        "argument1": "statusCode",
        "argument2Type": "String",
        "argument2": "ignorePart",
        "language": "JavaScript",
        "tagName": "_r3dir_encode_https"
    }
]


def _prepare_hackvertor_tags(main_domain: str):
    """Prepare Hackvertor tags for given main domain"""

    # JS code injection prevetion
    if re.fullmatch("[A-Za-z0-9-\.]+", main_domain) is None:
        main_domain = "r3dir.me"
        
    with open(hackvertor_encoder, 'r', encoding='utf-8') as file:
        encoder_content = file.read()
    
    for tag in tags:
        #Append encode/decode functions to each tag
        path_to_tag = os.path.join(hackvertor_tags_folder, tag["tagName"] + ".js")
        code_to_append = tag['code'].replace("[[MAIN_DOMAIN]]", main_domain)
        with open(path_to_tag, 'w', encoding='utf-8') as file:
            file.write(f"{encoder_content}\n{code_to_append}")
        #Saving PATH to the tag into Hackvertor json
        tag['code'] = path_to_tag
    return tags

def _cli():
    argParser = argparse.ArgumentParser(description='Encoded/decoder CLI tool for r3dir service')

    argParser.add_argument('-d', '--main_domain', type = str,
                            default = "r3dir.me",
                            help = f"Domain where r3dir tool is hosted on (default: %(default)s)")
    
    subparsers = argParser.add_subparsers(dest='mode', required=True, description="Encoding/decoding mode or Hackvertor tags generation")

    encoder = subparsers.add_parser('encode', help="r3dir CLI encoder")

    https_opts = encoder.add_mutually_exclusive_group()

    encoder.add_argument('target_url', type = str,
                            help = "Target URL which r3dir tool should redirect to")
    encoder.add_argument('-c', '--status_code', type = int,
                            default = 302,
                            help = f"HTTP status code of a redirect response (default: %(default)s)")
                       
    https_opts.add_argument('-i', '--ignore_part', type = str,
                            default = None,
                            help = f"String, which will be ignored during decoding. Used to bypass weak REGEXs")
    https_opts.add_argument("-s", "--https", help="HTTPS enforced encoding(TLS certificate length limitation)",
                            action="store_true")
    encoder.add_argument("--slient_mode", help="Slient mode for automations (e.g Hackvertor tags)",
                            action="store_true")

    decoder = subparsers.add_parser('decode', help="r3dir CLI decoder")        

    decoder.add_argument('encoded_domain', type = str,
                            help = "r3dir encoded domain to decode")
    
    hackvertor = subparsers.add_parser('hackvertor', help="Generate r3dir Hackvertor tags and copy them to clipboard")

    hackvertor.add_argument("--print", help="Output Hackvertor tags into terminal",
                            action="store_true")

    args = argParser.parse_args()

    if args.mode == 'encode':
        print(encode(args.target_url, args.status_code, args.main_domain, args.ignore_part, https_enforced = args.https, slient_mode=args.slient_mode))
    elif args.mode == 'decode':
        print(decode(args.encoded_domain, args.main_domain))
    elif args.mode == 'hackvertor':
        prepared_tags = _prepare_hackvertor_tags(args.main_domain)
        pyperclip.copy(json.dumps(prepared_tags))
        print("[+] Hackvertor tags were copied to a clipboard.")
        if args.print:
            pprint.pprint(prepared_tags)





class BaseCoderError(Exception):
    """Base class for r3dir encoding/decoding errors"""

class TooLongTarget(BaseCoderError):
    """Too long target to encode via domain-based redirection"""

class WrongEncodedURLFormat(BaseCoderError):
    """Wrong format of domain to decode"""

class Base32DecodingError(WrongEncodedURLFormat):
    """Base32-encoded target decoding error"""

class StatusCodeNotInRangeError(WrongEncodedURLFormat):
    """Status code is not in [200, 600) range"""


import pytest
from r3dir import encoder
from r3dir.exceptions import Base32DecodingError, StatusCodeNotInRangeError, WrongEncodedURLFormat, TooLongTarget

def test_valid_target():
    main_domain = "r3dir.me"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code

def test_normal_target_with_ignore_part():
    main_domain = "r3dir.me"
    ignore_part = "testingtest"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, ignore_part=ignore_part)
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code

def test_normal_target_with_ignore_part_and_https():
    main_domain = "r3dir.me"
    ignore_part = "testingtest"
    target, status_code = "http://169.254.169.254", 301
    with pytest.raises(TooLongTarget):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, ignore_part=ignore_part, https_enforced=True)

def test_normal_empty_encoded_target():
    main_domain = "r3dir.me"
    encoded_domain = "302.r3dir.me"
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert decoded_code == 302
    assert decoded_target == ""

def test_small_main_domain():
    main_domain = "a"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code

def test_large_main_domain():
    main_domain = "large.adfadf.af.adf.dsaf.ads.f.adsf.adsf.ads.f"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code

def test_large_main_domain_with_splitter():
    main_domain = "large.adfadf.af.adf.dsaf.ads.f.adsf.test.--.f"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code

def test_encoder_too_long_target():
    main_domain = "r3dir.me"
    target, status_code = "gopher://metadata.google.internal:80/xGET%20/computeMetadata/v1/instance/attributes/ssh-keys%20HTTP%2f%31%2e%31%0AHost:%20metadata.google.internal%0AAccept:%20%2a%2f%2a%0aMetadata-Flavor:%20Google%0d%0aTestHeader:%20Google", 301
    with pytest.raises(TooLongTarget):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, https_enforced=False)

def test_encoder_too_long_target_with_https_enforced():
    main_domain = "r3dir.me"
    target, status_code = "gopher://metadata.google.internal:80/xGET%20/computeMetadata/v1/instance/attributes/ssh-keys%20HTTP%2f%31%2e%31%0A", 301
    with pytest.raises(TooLongTarget):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, https_enforced=True)

def test_encoder_too_long_target_with_https_enforced_with_slient_mode():
    main_domain = "r3dir.me"
    target, status_code = "gopher://metadata.google.internal:80/xGET%20/computeMetadata/v1/instance/attributes/ssh-keys%20HTTP%2f%31%2e%31%0A", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, https_enforced=True, slient_mode=True)
    assert "too-long-target-" in encoded_domain

def test_encoder_too_long_target_with_slient_mode():
    main_domain = "r3dir.me"
    target, status_code = "gopher://metadata.google.internal:80/xGET%20/computeMetadata/v1/instance/attributes/ssh-keys%20HTTP%2f%31%2e%31%0AHost:%20metadata.google.internal%0AAccept:%20%2a%2f%2a%0aMetadata-Flavor:%20Google%0d%0aTestHeader:%20Google", 301
    encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain, https_enforced=False, slient_mode=True)
    assert "too-long-target-" in encoded_domain

def test_decoder_too_long_target():
    main_domain = "r3dir.me"
    encoded_domain =  "too-long-target-c1ee2cb84dee0891a3788b74cefdbb301ff8791f.301.r3dir.me"
    with pytest.raises(TooLongTarget):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_too_long_target_with_ignore_part():
    main_domain = "r3dir.me"
    encoded_domain =  "test.--.too-long-target-c1ee2cb84dee0891a3788b74cefdbb301ff8791f.301.r3dir.me"
    with pytest.raises(TooLongTarget):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_encoder_too_long_target_with_ignore_part():
    ignore_part = "testtest.test"
    main_domain = "r3dir.me"
    target, status_code = "gopher://metadata.google.internal:80/xGET%20/computeMetadata/v1/instance/attributes/ssh-keys%20HTTP%2f%31%2e%31%0AHost:%20metadata.google.internal%0AAccept:%20%2a%2f%2a%0aMetadata-Flavor:%20Google%0d%0a", 301
    with pytest.raises(TooLongTarget):
        encoded_domain = encoder.encode(target, status_code=status_code, ignore_part=ignore_part, main_domain=main_domain, https_enforced=False)

def test_decoder_status_code_out_of_range():
    main_domain = "r3dir.me"
    target, status_code = "http://169.254.169.254", 9999
    with pytest.raises(StatusCodeNotInRangeError):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_invalid_status_code():
    main_domain = "r3dir.me"
    target, status_code = "http://169.254.169.254", "string_status_code"
    with pytest.raises(WrongEncodedURLFormat):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_empty_status_code():
    main_domain = "r3dir.me"
    target, status_code = "http://169.254.169.254", ""
    with pytest.raises(WrongEncodedURLFormat):
        encoded_domain = encoder.encode(target, status_code=status_code, main_domain=main_domain)
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_invalid_encoded_target():
    main_domain = "r3dir.me"
    encoded_domain = "non-encoded.302.r3dir.me"
    with pytest.raises(Base32DecodingError):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    
def test_decoder_empty_encoded_target():
    main_domain = "r3dir.me"
    encoded_domain = ".302.r3dir.me"
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert decoded_target == ""

def test_decoder_empty_encoded_target_with_ignore_part():
    main_domain = "r3dir.me"
    encoded_domain = "test.--.302.r3dir.me"
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert decoded_target == ""

def test_decoder_malformed_ignore_part_splitter():
    main_domain = "r3dir.me"
    encoded_domain = "test.---.302.r3dir.me"
    with pytest.raises(Base32DecodingError):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_malformed_ignore_part_splitter_2():
    main_domain = "r3dir.me"
    encoded_domain = "test.--62epax5fhvj3zzmzie.302.r3dir.me"
    with pytest.raises(Base32DecodingError):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_malformed_ignore_part_splitter_3():
    main_domain = "r3dir.me"
    encoded_domain = "test.-.302.r3dir.me"
    with pytest.raises(Base32DecodingError):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_ignore_part_splitter_in_status_code():
    main_domain = "r3dir.me"
    encoded_domain = '62epax5fhvj3zzmzie.--.r3dir.me'
    with pytest.raises(WrongEncodedURLFormat):
        decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)

def test_decoder_splitted_encoded_subdomains():
    main_domain = "r3dir.me"
    target, status_code = "http://169.254.169.254", 301
    encoded_domain = '62epax5fhv.j3zzmzie.301.r3dir.me'
    decoded_target, decoded_code = encoder.decode(encoded_domain, main_domain=main_domain)
    assert target == decoded_target
    assert status_code == decoded_code


import math, base64, binascii
import unishox2
import hashlib

from .exceptions import Base32DecodingError, StatusCodeNotInRangeError, WrongEncodedURLFormat, TooLongTarget

MAX_DOMAIN_LENGTH = 253
MAX_SUBDOMAIN_LENGTH = 63
IGNORE_PART_SEP = "--"
MAX_COMPRESSION_TARGET_SIZE = 1024 # we assume that compression rate can't go futher than 85% on usefull links(1024*0.15*8/5=245,76)

def _chunkstring(string: str, length: int) -> tuple[str]:
    """Helper function, which split string in several chunks not great than indicated lenght"""

    return (string[0+i:length+i] for i in range(0, len(string), length))

def _b32encode_dns(string: str) -> tuple:
    """Base32 encoding of string with Unishox2 compression. Accept string to encode.
       Returns tuple of chunked encoded data, spllited to be used as subdomains"""

    data, _ = unishox2.compress(string)
    encoded_string = base64.b32encode(data).decode('ascii')

    #stripping '=' to save characters in subdomains
    encoded_string = encoded_string.rstrip("=").lower()
    #splitting in chunks due to subdomain length limitations
    subdomains = _chunkstring(encoded_string, MAX_SUBDOMAIN_LENGTH)
    return tuple(subdomains)

def _b32decode_dns(subdomains: tuple | list) -> str:
    """Base32 decoding with unishox2 compression. Accept tuple of subdomains,
       containing encoded string. Returns decoded and decompresed string"""
    
    encoded_string = "".join(subdomains)

    #padding Base32 for error-less decoding
    pad_length = math.ceil(len(encoded_string) / 8) * 8 - len(encoded_string)
    encoded_target = encoded_string + "=" * pad_length

    decoded_data = base64.b32decode(encoded_target, casefold = True)
    decompresed_data = unishox2.decompress(decoded_data, MAX_COMPRESSION_TARGET_SIZE)
    return decompresed_data

def _is_too_long_target_error(encoded_subdomains: tuple | list):
    """Detects targets, which was too long to encode but tool was used in """
    for domain in encoded_subdomains:
        if domain.startswith("too-long-target-"):
            target_url_hash = domain.replace("too-long-target-", "") 
            return target_url_hash
    return False

def encode(target: str, status_code: int, main_domain: str, 
            ignore_part: str | None = None, https_enforced: bool = False, slient_mode: bool = False) -> str:
    """"r3dir encoder method. Accepts redirection target, status code of redirect
        and main_domain of used redirection server. Returns domain with encoded target.
        For optional parameters description, reference CLI tool help."""

    subdomains = ".".join(_b32encode_dns(target))
    
    encoded_domain = f"{subdomains}.{status_code}.{main_domain}"

    #check whether encoded domain length corresponds to limitations of DNS and TLS certificates
    try:
        if https_enforced:
            if ignore_part or len(subdomains) > MAX_SUBDOMAIN_LENGTH:
                raise TooLongTarget(f"The target length is longer than maximum allowed for HTTPS mode. Remove ignoring part, or short the target.")
        elif ignore_part:
            encoded_domain = f"{ignore_part}.{IGNORE_PART_SEP}.{encoded_domain}"

        if len(encoded_domain) > MAX_DOMAIN_LENGTH:
            raise TooLongTarget(f"The target length is longer than maximum allowed. Remove ignoring part or short the target.")
    except TooLongTarget:
        if not slient_mode:
            raise
        target_hash = hashlib.sha1(target.encode("UTF-8")).hexdigest()
        error_domain = f"too-long-target-{target_hash}.{status_code}.{main_domain}"
        return error_domain
    
    return encoded_domain

def decode(domain: str, main_domain: str) -> tuple[str, int]:
    """"r3dir decoder method. Accepts encoded domain and main domain of the redirection server.
        Returns redirection target and status code for redirect responce"""

    subdomains = domain.split('.')

    main_domain_chunk_count = len(main_domain.split('.'))

    subdomains_without_main = subdomains[:-main_domain_chunk_count]
    
    start_of_encoded_target = 0
    if IGNORE_PART_SEP in subdomains_without_main:
        start_of_encoded_target = subdomains_without_main.index(IGNORE_PART_SEP) + 1
    
    encoded_subdomains = subdomains_without_main[start_of_encoded_target:]

    if target_url_hash := _is_too_long_target_error(encoded_subdomains):
        raise TooLongTarget(f"The target length has been too long for encoder. Target's SHA-1: {target_url_hash}")

    try:
        status_code = int(encoded_subdomains[-1])
    except (IndexError, ValueError) as e:
        raise WrongEncodedURLFormat("Can't read status code.")
    if status_code not in range(200, 600):
        raise StatusCodeNotInRangeError("Status code is not in [200, 600) range")
    
    try:
        target = _b32decode_dns(encoded_subdomains[:-1])
    except (UnicodeDecodeError, binascii.Error) as e:
        raise Base32DecodingError("Base32-encoded target decoding error")
    
    return target, status_code

